# <path to game directory>/addons/eventscripts/_engines/python/Lib/esutils/
# tools.py
# by Adam Cunnington

"""Provide pythonic, intuitive helper functions (and wrappers) to be commonly 
used by both libraries and modules.

"""

import operator
import string

import es

__all__ = (
    "ToolsError", 
    "GAME_DIR", 
    "Filter", 
    "echo", 
    "format_verbose_name", 
    "get_user_id_from_handle", 
    "get_user_id_from_name", 
    "get_user_id_from_steam_id", 
    "is_connected", 
    "message", 
    "output", 
    "tell", 
)

_TEAM_COLOUR = "teamcolour"

_COLOURS = {
    "darkgreen": "\x05",
    "green": "\x04",
    "lightgreen": "\x03",
    "yellow": "\x01",
    _TEAM_COLOUR: "\x03"
}

_DELIMITER = "$"

_EXISTS_USER_ID = "userid"

_USERMSG_ECHO = 2
_USERMSG_SAYTEXT2 = 1

GAME_DIR = es.ServerVar("eventscripts_gamedir")

class ToolsError(Exception):
    """General error encountered relating to esutils.tools"""
    
class Filter(object):
    """Create a filter object that can be further filtered by the object's 
    attribute values.
    
    """
    
    def __init__(self, original_objects, current_objects=None, attribute=""):
        """Instantiate a Filter object.
        
        Arguments:
        original_objects - the original set of objects which all other related 
        filter objects are filtered from.
        current_objects (Keyword Default: None) - the current set of objects 
        that have not been filtered out.
        attribute (Keyword Default: "") - the name of the attribute for 
        arithmetical / logical comparison operators to use.
        
        """
        self._original_objects = set(original_objects)
        if current_objects is None:
            self._current_objects = original_objects
        else:
            self._current_objects = current_objects
        self._attribute = attribute
        
    def __contains__(self, value):
        return Filter(self._original_objects, 
                      self._get_objects(operator.contains, value), 
                      self._attribute)
                      
    def __eq__(self, value):
        return Filter(self._original_objects, 
                      self._get_objects(operator.eq, value), self._attribute)
                      
    def __ge__(self, value):
        return Filter(self._original_objects, 
                      self._get_objects(operator.ge, value), self._attribute)
                      
    def __getattr__(self, name):
        return Filter(self._original_objects, self._current_objects, name)
        
    def __gt__(self, value):
        return Filter(self._original_objects, 
                      self._get_objects(operator.gt, value), self._attribute)
                      
    def __inv__(self):
        return Filter(self._original_objects, self._original_objects - 
                       self._current_objects, self._attribute)
                       
    def __iter__(self):
        return self._current_objects.__iter__()
        
    def __le__(self, value):
        return Filter(self._original_objects, 
                      self._get_objects(operator.le, value), self._attribute)
                      
    def __lt__(self, value):
        return Filter(self._original_objects, 
                      self._get_objects(operator.lt, value), self._attribute)
                      
    def __ne__(self, value):
        return Filter(self._original_objects, 
                      self._get_objects(operator.ne, value), self._attribute)
                      
        def _get_objects(self, operation, value):
            for current_object in self._current_objects:
                if operation(getattr(current_object, self._attribute), value):
                    yield current_object
                    
    def filter(self, *args):
        """Return a new filter object based on a number of comparisons.
        
        Arguments:
        args - the comparisons to determine which objects should be filtered 
        out in order to leave a set of objects for the new filter object.
        
        """
        return Filter(self._original_objects, set(itertools.chain(*args)))
        
def _get_colour(index):
    colour = es.getindexprop(index, "CBaseEntity.m_clrRender")
    return (colour & 0xff, (colour & 0xff00) >> 8, 
            (colour & 0xff0000) >> 16, (colour & 0xff000000) >> 24)
            
def _get_message(message):
    template = string.Template(message)
    template.delimiter = _DELIMITER
    return template.substitute(_COLOURS)
    
def _set_colour(index, value):
    es.setindexprop(index, "CBaseEntity.m_nRenderMode", 
                    es.getindexprop(index, "CBaseEntity.m_nRenderMode") | 1)
    es.setindexprop(index, "CBaseEntity.m_nRenderFX", 
                    es.getindexprop(index, "CBaseEntity.m_nRenderFX") | 256)
    red, green, blue, alpha = value
    colour = red + (green << 8) + (blue << 16) + (alpha << 24)
    if colour >= 2 ** 31: 
        colour -= 2 ** 32
    es.setindexprop(index, "CBaseEntity.m_clrRender", colour)
    
def echo(user_id, text=" "):
    """Message a single user through client console. If text is omitted, a 
    echo a blank line.
    
    Arguments:
    user_id - the user ID to send the message to.
    text (Keyword Default: " ") - the text to be output.
    
    """
    es.usermsg("create", "echo", "TextMsg")
    es.usermsg("write", "byte", "echo", _USERMSG_ECHO)
    es.usermsg("write", "string", "echo", text)
    
def format_verbose_name(basename):
    """Format verbose name from basename by replacing underscores with spaces 
    and then titling. Return the result.
    
    Arguments:
    basename - the basename to format the verbose name from.
    
    """
    return basename.replace("_", " ").title()
    
def get_user_id_from_handle(handle):
    """Return the user's unique session ID from a handle ID.
    
    Arguments:
    handle - the handle ID to get a user's unique session ID from.
    
    """
    for user_id in es.getUseridList():
        if steam_id == es.getplayerhandle(user_id):
            return user_id
    raise KeyError(handle)
    
def get_user_id_from_name(name):
    """Return the user's unique session ID from a name. Full-name matches have 
    a higher order of precendence than partial matches.
    
    Arguments:
    name - the name to get a user's unique session ID from.
    
    """
    for user_id in es.getUseridList():
        user_name = es.getplayername(user_id)
        if name == user_name:
            return user_id
        if name not in user_name:
            continue
        # Only interested in the first partial match.
        return user_id
    raise KeyError(name)
    
def get_user_id_from_steam_id(steam_id):
    """Return the user's unique session ID from a steam ID.
    
    Arguments:
    steam_id - the steam ID to get a user's unique session ID from.
    
    """
    for user_id in es.getUseridList():
        if steam_id == es.getplayersteamid(user_id):
            return user_id
    raise KeyError(steam_id)
    
def is_connected(user_id):
    """Return whether or not a given unique session ID is active.
    
    Arguments:
    user_id - the unique session ID of the user.
    
    """
    return bool(es.exists(_EXISTS_USER_ID, user_id))
    
def message(text):
    """Announce formatted message to user using es.msg with prefixed addon 
    information.
    
    Arguments:
    text - the text to be output.
    
    """
    es.msg("#multi", _get_message(text))
    
def output(text):
    """Output text to the server console.
    
    Arguments:
    text - the text to be output.
    
    """
    es.dbgmsg(0, text)
    
def tell(user_id, text, index=None):
    """Message a single user through say-chat. If no index is explicitly 
    passed, get the index of the user that is to be messaged.
    
    Arguments:
    user_id - the user ID to send the message to.
    text - the text to be output.
    index (Keyword Default: None) - the index to be used for the usermsg 
    formatting.
    
    """
    if _DELIMITER + _TEAM_COLOUR in text:
        if index is None:
            index = es.getindexfromhandle(es.getplayerhandle(userid))
        es.usermsg("create", "saytext2", "SayText2")
        es.usermsg("write", "byte", "saytext2", index)
        es.usermsg("write", "byte", "saytext2", _USERMSG_SAYTEXT2)
        es.usermsg("write", "string", "saytext2", _get_message(text))
        es.usermsg("send", "saytext2", user_id)
        es.usermsg("delete", "saytext2")
        return
    es.tell(user_id, "#multi", _get_message(text))